(* 
   SNEG - Mathematica package for calculations with non-commuting
   operators of the second quantization algebra
   
   Copyright (C) 2006 Rok Zitko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

   Contact information:
   Rok Zitko
   F5 - Condensed matter physics
   "Jozef Stefan" Institute
   Jamova 39
   SI-1000 Ljubljana
   Slovenia

   rok.zitko@ijs.si

   $Id: sneg-Flat-OneIdentity.m,v 1.1 2006/07/04 16:13:22 rok Exp rok $
*)

BeginPackage["Sneg`"];

Print["sneg $Id: sneg-Flat-OneIdentity.m,v 1.1 2006/07/04 16:13:22 rok Exp rok $"];
Print["Copyright (C) 2006 Rok Zitko, rok.zitko@ijs.si"];
Print["Rok Zitko, F5 - Condensed matter physics, \"Jozef Stefan\" Institute,
Jamova 39, SI-1000 Ljubljana, Slovenia"];
Print["SNEG comes with ABSOLUTELY NO WARRANTY; for details
type snegwarranty[].  This is free software, and you are welcome
to redistribute it under certain conditions; type snegconditions[] 
for details."];

$SnegVersion = "1.55";


(* TO DO *)
(* ===== *)
(* How to write sum[] simplification rules to be more deterministic? *)
(* How to extend Mathematica's Simplify[] function?? *)
(* Majorana Fermions *)
(* Bosons *)
(* - Make functions Listable *)
(* - Make nc OneIdentity -> this should be thoroughly tested!! *)
(* - Make nc Flat *)
(* - If Debugging mode off, prevent multiple loadings of this package *)
(* - Put non-exportable functions in sub-context 'Private' *)

(* 
A few conventions:
  z - general (complex) constant
  x - general constant, known to be real
  op - operators
  a,b,c - general expressions
  i - first index == type, 1 for annihilation, 0 for creation
  j - other indeces, site, spin etc.
  k - operator indeces, including the first (type) index
    - an exception are SEA ordered operators, where k is used to
      denote the second (wavenumber) index [for convenience]
  sigma - spin, UP or DO
  v - state in number representation

  l - list

  i - integer

Every operator has at least one index, which specifies whether it
is a creation or annihilation operator. Therefore a generic pattern
is op_[i_, j___]. Moreover, if the operator contains a spin index,
it is always the last index!
*)



(**** Usage messages ****)
snegwarranty::usage =
  "snegwarranty[] prints out the warranty";
snegconditions::usage =
  "snegconditions[] prints out the confitions";

snegfermionoperators::usage = 
  "snegfermionoperators[a, b, c] defines a, b, and c to be 
   fermionic operators. They obey canonical anti-commutation relations.";
snegrealconstants::usage =
  "snegrealconstants[x, y] defines x and y to be real constants.
   Constants are factored out from operator strings within nc[] blocks.
   They are invariant under the conjugation.";
snegcomplexconstants::usage =
  "snegcomplexconstants[z, w] defines z and w to be complex constants.
   Constants are factored out from operator strings within nc[] blocks.";
snegfreeindexes::usage =
  "snegfreeindexes[k, sigma] defines k and sigma to be indexes. This is
   taken into account when simplifying sum[] blocks.";
nc::usage = 
  "nc[...] is a string of operators that do not (necessarily) commute,
   i.e. a non-commutative multiplication.";
operatorQ::usage =
  "operatorQ[c] returns True if c is an operator.";
ordering::usage = 
  "ordering[c] = EMPTY | SEA defines the vacuum that corresponds to
   an operator. EMPTY corresponds to an empty state vacuum, while
   SEA denotes a Fermi sea filled up to a Fermi level.";
snegOrderedQ::usage =
  "snegOrderedQ[c, {i1, i2}] returns True if operators c[i1] and c[i2]
   are canonically ordered. A predefined functions exist for ordering[c]
   = EMPTY | SEA, otherwise this function must be provided by the user.";
overrideSEA::usage =
  "overrideSEA[expr] reorders an expression, disregarding the difference
   between particle- and hole-creation operators. The resulting expression
   will not be normal ordered with respect to the Fermi sea vacuum.";
acmt::usage = 
  "acmt[op, {i1}, {i2}] defined the value of the anti-commutator
   between op[i1] and op[i2]. By default, the value is given by the
   canonical anti-commutator for fermionic operators. For indexed
   operators, a user-provided overloaded function can be provided.";
komutator::usage =
  "komutator[a, b] = nc[a, b] - nc[b,a]";
antikomutator::usage =
  "antikomutator[a, b] = nc[a, b] + nc[b, a]";
dd::usage = 
  "dd[expr] denotes the double dots surrounding an expresions, i.e.
   a normal ordered string";
normalorder::usage =
  "normalorder[expr] normal orders an expression";
normalorderwick::usage =
  "normalorderwick[expr} normal orders an expression using the Wick's theorem";
conj::usage =
  "conj[expr] conjugates the expression 'expr'.";
sum::usage =
  "sum[expr, {indeces}] denotes a symbolic sum of expression 'expr'
   over indeces 'indeces'.";
sumThread::usage =
  "sumThread is a rule for threading a sum[{expr1, expr2}, {indeces}]
   over the list in the first argument.";
sumExpand::usage =
  "sumExpand is a rule for expanding sums over the expression in the first
   argument.";
sumCollect::usage = 
  "sumCollect is a rule for collecting sums over the expression in the
   first argument.";
sumSimplify::usage =
  "sumSimplify is a rule for simplifying sums over the expression in the
   first argument.";
sumAbstractIndex::usage =
  "sumAbstractIndex[expr] replaces sum indeces with abstract indeces.
   This is useful for simplifications in the case of several equivalent
   summation indeces.";
sumNameIndex::usage =
  "sumNameIndex[expr] is to be used in conjunction with 
   sumAbstractIndex[expr].";
sumFullSimplify::usage =
  "sumFullSimplify[expr] attempts to simplify an expression involving
   symbolic sums. A heuristic approach is used, which usually does a good
   job, but some further simplifying by hand might sometimes be necessary.";
vev::usage =
  "vev[expr] returns the vacuum expectation value of an operator expression.";
onvac::usage =
  "onvac[expr] drops vacuum-annihilating operators from an operator 
   expression.";
mdot::usage =
  "mdot[a, b] returns the inner product between a and b using 
   nc[] multiplication.";
inner::usage =
  "inner[a, b] returns the inner product between a and b using
   nc[] multiplication.";
outer::usage =
  "outer[a, b] returns the outer (direct) product between a and b
   using nc[] multiplication.";
VMV::usage =
  "VMV[v1, m, v2] computes the vector-matrix-vector using nc[] inner
   product multiplication.";
number::usage =
  "number[c[i]] returns the number operator corresponding to the operator
   c[i]. number[c[i], sigma] with sigma=UP | DO returns the number operator
   for a given spin projection only.";
hubbard::usage =
  "hubbard[c[i]] returns the local (Hubbard) electron-electron repulsion
   operator corresponding to operator c[i].";
hamiltonian::usage =
  "hamiltonian[type, ...] is a stub: it returns the Hamiltonian 'type'.";
chargecharge::usage =
  "chargecharge[c[i], d[j]] returns the charge-charge repulsion operator
   corresponding to operators (sites) c[i] and d[j].";
PauliX::usage =
  "Pauli's X matrix";
PauliY::usage =
  "Pauli's Y matrix";
PauliZ::usage =
  "Pauli's Z matrix";
PauliPlus::usage =
  "Pauli's + (spin raising) matrix";
PauliMinus::usage =
  "Pauli's - (spin lowering) matrix";
spinxyzgen::usage =
  "spinxyzgen[g] generates x, y, and z components of a S=1/2 spin
   operator corresponding to an (operator) expression g with a free
   spin index 'sigma'. For example, the convention spin operators are 
   obtained by spinxyzgen[op[CR, sigma]].";
spinxyz::usage =
  "spinxyz[op[j]] returns the x, y, and z components of a S=1/2 spin
   operator corresponding to a fermionic S=1/2 operator op[j].";
spinall::usage =
  "spinall[op[j]] returns the x, y, and z components of a S=1/2 spin
   operator, as well as the S^2 operator, corresponding to a fermionic
   operator op[j].";
spinss::usage =
  "spinss[op[j]] returns the S^2 total S=1/2 spin operator squared that
   corresponds to a fermionic operator op[j].";
spinx::usage =
  "spinx[op[j]] returns the x component of a S=1/2 spin operator that
   corresponds to a fermionic operator op[j].";
spiny::usage =
  "spiny[op[j]] returns the y component of a S=1/2 spin operator that
   corresponds to a fermionic operator op[j].";
spinz::usage =
  "spinz[op[j]] returns the z component of a S=1/2 spin operator that
   corresponds to a fermionic operator op[j].";
spinplus::usage =
  "spinplus[op[j]] returns the S=1/2 spin raising operator that
   corresponds to a fermionic operator op[j].";
spinminus::usage =
  "spinplus[op[j]] returns the S=1/2 spin lowering operator that
   corresponds to a fermionic operator op[j].";
spinspin::usage =
  "spinspin[a[i], b[j]] returns the spin-spin scalar product between
   two S=1/2 spin operators corresponding to fermionic operators a[i]
   and b[j].";
spinspinpm::usage =
  "spinspinpm[a[i], b[j]] returns the +- part of the scalar product
   between two S=1/2 spin operators corresponding to fermionic operators 
   a[i] and b[j].";
spinspinmp::usage =
  "spinspinmp[a[i], b[j]] returns the -+ part of the scalar product
   between two S=1/2 spin operators corresponding to fermionic operators 
   a[i] and b[j].";
spinspinxy::usage =
  "spinspinxy[a[i], b[j]] returns the transverse part of the scalar product
   between two S=1/2 spin operators corresponding to fermionic operators 
   a[i] and b[j].";
spinspinz::usage =
  "spinspinxy[a[i], b[j]] returns the longitudinal part of the scalar product
   between two S=1/2 spin operators corresponding to fermionic operators 
   a[i] and b[j].";
manyspin::usage =
  "manyspin[{a[i], b[j],...}] returns the x, y, and z components, as
   well as the total spin operator squared, corresponding to a list
   of operators given as the argument.";
halfintegerQ::usage =
  "halfintegerQ[z] returns True if z is an integer or a half-integer.";
spinmatrixX::usage =
  "spinmatrixX[S] returns the matrix of the X component of the spin
   operator for arbitrary spin S.";
spinmatrixY::usage =
  "spinmatrixY[S] returns the matrix of the Y component of the spin
   operator for arbitrary spin S.";
spinmatrixZ::usage =
  "spinmatrixZ[S] returns the matrix of the Z component of the spin
   operator for arbitrary spin S.";
spinmatrixP::usage =
  "spinmatrixP[S] returns the matrix of the spin-raising operator 
   for arbitrary spin S.";
spinmatrixM::usage =
  "spinmatrixM[S] returns the matrix of the spin-lowering operator 
   for arbitrary spin S.";
direct::usage =
  "direct[b1, b2] returns a direct product of two sets of operator
   expressions (nc[] multiplication is used.";
nambu::usage =
  "nambu[c[i], nr] returns the Nambu-spinor components corresponding
   to a fermionic operator c[i]. The phase is (-1)^nr. The default
   value of nr is 0.";
isospin::usage =
  "isospin[c[i], nr] returns the isospin operator corresponding
   to a fermionic operator c[i]. The phase is (-1)^nr. The default 
   value of nr is 0.";
manyisospin::usage =
  "manyisospin[{ops}, {phases}] returns the X, Y, and Z components
   of the total isospin operator for a list of fermionic operators.
   The phases of operators must be given as the second argument.
   The total isospin squared operator is also returned.";
basis::usage =
  "basis[op[i]] returns the basis states (creation operators) for
   a single site, i.e. for operator op[i].";
pow::usage =
  "pow[expr, i] raises expression expr to i-th power using the nc[]
   multiplication.";
hop::usage =
  "hop[a[i], b[j]] returns the electron hopping operator between
   sites a[i] and b[j]. hop[a[i], b[j], sigma] with sigma=UP | DO
   does the same for a single spin projection sigma.";
spinfliphop::usage =
  "spinfliphop[a[i], b[j]] returns the electron hopping with spin flip
   operator between sites a[i] and b[j].";
holehop::usage = 
  "holehop[h[i], p[j]] returns the electron hopping operator for
   the case of a hole operator h[i] and particle operator p[j].";
twohop::usage =
  "twohop[a[i], b[j]] returns the two-particle hopping operator
   between sites a[i] and b[j].";
invertspin::usage =
  "invertspin[expr] inverts the spin of the operator expression expr.
   The spin index must be the second index of each operator appearing
   in the expression.";
ruleCollect::usage =
  "ruleCollect is a rule for collecting prefactors of nc[] operator 
   strings.";
makebasis::usage =
  "makebasis[ops] constructs a basis for enlisted operators.
   Example: makebasis[{a[],f[1]}].
   Expansion over both spin orientations is automatically performed.";
goodopQ::usage =
  "goodopQ[c[i]] returns True if the operator c[i] appears in the basis.
   The type of the operator (CR vs. AN) is disregarded.";
op2ndx::usage =
  "op2ndx[c[i]] returns the index of operator c[i] in the basis.";
ndx2op::usage =
  "ndx2op[ndx] returns the basis operator corresponding to index ndx.";
vacuum::usage =
  "vacuum[] returns the vacuum vector in vc[] representation for a
   basis that was previously defined using function makebasis[].";
ap::usage = 
  "ap[op, vc] applies an operator op to a vector vc in the occupation
   number representation.";
vc2ops::usage =
  "vc2ops[expr] transforms vectors in the occupation-number representation
   to the corresponding operator strings in expression 'expr'.";
vc2opsmap::usage =
  "vc2opsmap[list] applies vc2ops to each element of list.";
scalarproduct::usage =
  "scalarproduct[a, b] computes the scalar product between two expressions
   consisting of vectors in the occupation number representation.";
scalarproductop::usage =
  "scalarproductop[a, b] computes the vacuum expectation value of the
   product between operator expressions a and b. The expression on the
   left, a, is automatically conjugated.";
expvop::usage = 
  "expvop[op, cr] computes the bracket <cr|op|cr>, where op is an
   operator expression and cr is some state (in the creation operator
   representation).";
matrixrepresentationvc::usage =
  "matrixrepresentationvc[op, basis] computes the matrix representation
   of the operator op in a given basis (occupation number representation).
   matrixrepresentationvc[op, basisL, basisR] returns the matrix 
   representation of an operator in the case of different basis for 
   left-multiplication and right-multiplication.";
matrixrepresentationop::usage = 
  "matrixrepresentationop[op, basis] computes the matrix representation
   of the operator op in a given basis (creation operator representation).";
qszbasisvc::usage =
  "qszbasisvc[{ops}] returns the basis with well defined charge and
   spin projection quantum numbers (Q,S_z). An appropriate occupation
   number basis for operators {ops} must be defined using makebasis[]
   before calling qszbasisvc[].";
qszbasis::usage =
  "See qszbasisvc[].";
interleave::usage =
  "interleave[{a1, a2, ...}, {b1, b2, ..}] returns {a1, b1, a2, b2,...}.";
qsbasis::usage =
  "qsbasis[{ops}] returns the basis with well defined charge and
   total spin quantum numbers (Q,S).";
zeroonvac::usage =
  "zeroonvac[expr] drops vacuum-annihilating parts of expression expr.";
VACUUM::usage =
  "VACUUM is a placeholder for vacuum in operator expressions. If an
   annihilation operator stands just before VACUUM, the expression is zero.
   Likewise, if a creation operators stands just after conj[VACUUM], the
   expression drops. VACUUM is normalized to 1, i.e. 
   vev[nc[conj[VACUUM], VACUUM]] = 1.";
keyselect::usage =
  "keyselect[list, key] selects sublists with the first element equal to key.";
keymerge::usage =
  "keymerge[list1, list2] merges two lists, so that the sublists with
   the same first element are combined.";
CR::usage =
  "c[CR] is a creation operator.";
AN::usage =
  "c[AN] is an annihilation operator.";
DO::usage =
  "c[CR, DO] is a spin-down (creation) operator.";
UP::usage =
  "c[CR, UP] is a spin-up (creation) operator.";
PROJ0::usage =
  "Argument for projector[]. Subspace with zero occupancy.";
PROJUP::usage =
  "Argument for projector[]. Subspace with spin up particle.";
PROJDO::usage =
  "Argument for projector[]. Subspace with spin down particle.";
PROJ2::usage =
  "Argument for projector[]. Subspace with double occupancy.";
PROJ1::usage =
  "Argument for projector[]. Subspace with single occupancy.";
PROJ02::usage =
  "Argument for projector[]. Subspace with either zero or double occupancy.";
projector::usage =
  "projector[op[j], PROJ] returns the projection operator corresponding
   to the fermionic operator op[j]. PROJ = PROJ0 | PROJUP | PROJDO
   | PROJ2 | PROJ1 | PROJ02.";
projector0::usage =
  "projector0[op[j]] returns the projection operator corresponding to the
   fermionic operator op[j]. It projects to the subspace with zero
   occupancy.";
projector0::usage =
  "projector0[op[j]] returns the projection operator corresponding to the
   fermionic operator op[j]. It projects to the subspace with zero
   occupancy.";
projectorUP::usage =
  "projectorUP[op[j]] returns the projection operator corresponding to the
   fermionic operator op[j]. It projects to the subspace a single
   spin up particle.";
projectorDO::usage =
  "projectorDO[op[j]] returns the projection operator corresponding to the
   fermionic operator op[j]. It projects to the subspace a single
   spin down particle.";
projector2::usage =
  "projector2[op[j]] returns the projection operator corresponding to the
   fermionic operator op[j]. It projects to the subspace with double
   occupancy.";
projector1::usage =
  "projector1[op[j]] returns the projection operator corresponding to the
   fermionic operator op[j]. It projects to the subspace with single
   occupancy.";
projector02::usage =
  "projector02[op[j]] returns the projection operator corresponding to the
   fermionic operator op[j]. It projects to the subspace with either zero
   or double occupancy.";
decomposeop::usage =
  "decomposeop[vec, basis] decomposes a vector into components with
   respect to the given basis in the creation operator representation.";
decomposevc::usage =
  "decomposevc[vec, basis] decomposes a vector into components with     
   respect to the given basis in the occupation number representation.";
orthogop::usage =
  "orthogop[vecs, basis] orthogonalizes a list of vectors vecs in the
   creation operator representation basis 'basis'.";
orthogvc::usage =
  "orthogvc[vecs, basis] orthogonalizes a list of vectors vecs in the
   occupation number representation basis 'basis'.";
transformQStoIS::usage =
  "transformQStoIS[basis] transforms from a basis with well defined
   (Q,S) quantum numbers to a basis with well defined (I,S) quantum numbers.
   basis is a basis list in the format as returned by function qsbasis[].
   Prior to calling transformQStoIS, Tminus (isospin lowering) operator must
   be defined.";
bra::usage =
  "bra[i] denotes the Dirac bra |i>.";
ket::usage =
  "ket[i] denotes the Dirac ket <i|.";
braketrule::usage =
  "braketrule[bra[b], ket[k]] defines the value of the braket <b|k>.";
spinket::usage =
  "spinket[S] returns the spin basis kets for spin S.";
spinbra::usage =
  "spinbra[S] returns the spin basis bras for spin S.";
spinketbraX::usage =
  "spinketbraX[S] returns the X-component of the spin operator for spin S
   in the bra-ket representation.";
spinketbraY::usage =
  "spinketbraY[S] returns the Y-component of the spin operator for spin S
   in the bra-ket representation.";
spinketbraZ::usage =
  "spinketbraZ[S] returns the Z-component of the spin operator for spin S
   in the bra-ket representation.";
spinketbraP::usage =
  "spinketbraP[S] returns the spin lowering operator for spin S
   in the bra-ket representation.";
spinketbraM::usage =
  "spinketbraM[S] returns the spin raising operator for spin S
   in the bra-ket representation.";
spinbasis::usage =
  "spinbasis[S] returns the basis for spin S in the ket representation.";
transformfunc::usage =
  "transformfunc[list, rule] applies a rule to the states 'list' in the
   occupation number representation and orthogonalizes the result.";
transformbasis::usage =
  "transformbasis[basis, rule] applies a rule to the states in the
   basis (occupation number representation) and orthogonalizes the result.";
mergebasis::usage =
  "mergebasis[{basis1, basis2,...}] merges several sets of basis states.";
applybasis::usage =
  "applybasis[b, fn] applies function fn to each basis state of the
   basis b.";
QSpseudospin::usage =
  "TO DO";
transformQStoPH::usage =
  "transformQStoPH[b, Nph] generates a direct product of a basis b for
   fermions with a basis with up-to Nph excited local phonons, i.e. Nph
   is the phonon number cutoff.";
phononnumber::usage =
  "phononnumber[Nph] returns the phonon number operator in the bra-ket
   representation. Nph is the phonon number cutoff.";
phononplus::usage =
  "phononplus[Nph] returns the phonon raising operator in the bra-ket
   representation. Nph is the phonon number cutoff.";
phononminus::usage = 
  "phononminus[Nph] returns the phonon lowering operator in the bra-ket
   representation. Nph is the phonon number cutoff.";
orthogbasisop::usage =
  "orthogbasisop[bz1, bz0] orthogonalizes the basis bz1 with respect
   to the original basis bz. Both basis sets must be given in the creation 
   operator representation.";
orthogbasisvc::usage =
  "orthogbasisvc[bz1, bz0] orthogonalizes the basis bz1 with respect
   to the original basis bz. Both basis sets must be given in the occupation 
   number representation.";
transformtoLR::usage =
  "transformLR[b, map] transforms the basis b to a basis with well
   defined parity. map is the list of the operators so that the sites
   are symmetric with respect to the middle of this list.";
snegold2newrules::usage =
  "Given a set of old operators o, a set of new operators n and
   a set of operator expressions for the new basis, snegold2newrules[o, n, r]
   produces the transformation rules from the old basis to the new basis.";
isCR::usage = 
  "isCR[op[j]] returns 1 if op[j] is a cration operator and 0 otherwise.";
icAN::usage =
  "isAN[op[j]] returns 0 if op[j] is an annihilation operator and 0
   otherwise.";
contraction::usage =
  "contraction[op1, op2] contracts op1 and op2 according to the
   standard definition.";
wick::usage =
  "wick[expr] returns an operator string expressed using normal ordered
   strings and contractions. This is an application of Wick's theorem.";
vevwick::usage =
  "vevwick[expr] calculates the vacuum expectation value of an operator
   string using Wick's theorem.";


(** License related functions **)
snegwarranty[] := Print["NO WARRANTY\n\n
BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION."];

snegconditions[] := Print[
"IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES."];



(** Load required packages **)
(* Used in orthogop[] and orthogvc[] *)
Needs["LinearAlgebra`Orthogonalization`"];
(* Used in contractedpairs[] *)
Needs["DiscreteMath`Combinatorica`"];



(* Standard shorthands for operator indeces; some functions depend
explicitly upon these particular values, notably ap[] ... For this
reasion, these values have Protected attribute. *)

ClearAttributes[{CR, AN, DO, UP}, Protected];

CR = 0;
AN = 1;
DO = 0;
UP = 1;

SetAttributes[{CR, AN, DO, UP}, Protected];

(* Define which symbols denote fermionic operators *)
snegfermionoperators[l__] := Scan[
  {operatorQ[#] = True;
   SetAttributes[#, NHoldAll];
   fermionQ[#] = True;}&, 
{l}];

(* Define which symbols are numerical constants *)
snegrealconstants[l__] := Scan[
  {NumericQ[#] = True; 
   Conjugate[#] ^= #;}&, 
{l}];
snegcomplexconstants[l__] := Scan[
  {NumericQ[#] = True;}&, 
{l}];

(* Define which symbols are free indexes that appear in sum[]s *)
snegfreeindexes[l__] := Scan[
  {NumericQ[#] = True; (* Added 3/7/06 *)
   freeindex[#] = True;}&, 
{l}];

(* Make commonly used functions Numeric *)
SetAttributes[KroneckerDelta, NumericFunction];
SetAttributes[UnitStep, NumericFunction];

(* Checking for operator quantities with indeces*)
operatorQ[op_?operatorQ[k__]] := True;

(* Make "op" be a linear operator in all of its arguments. *)
(* Such operators are, for example, nc[], dd[], ap[], invertspin[]... *)

sneglinearoperator[op_] := {
  op[a___, z_?NumericQ, b___] := z op[a, b];
  op[a___, z_?NumericQ b_, c___] := z op[a, b, c];
  op[a___, b1_ + b2_, c___] := op[a, b1, c] + op[a, b2, c];
};

(* Similar to sneglinearoperator[op], but the operator op is
   made linear in the first argument only. Such operators are,
   for example, vev[], onvac[], zeroonvac[].*)

sneglinearoperatorFirst[op_] := {
  op[z_?NumericQ] := z;
  op[z_?NumericQ, b__] := z op[b];
  op[z_?NumericQ b_, c___] := z op[b, c];
  op[b1_ + b2_, c___] := op[b1, c] + op[b2, c];
};

(* nc is a reserved word and it is obviously never an operator. 
This definition speeds up some tests. *)
operatorQ[nc] = False; 

(**** Define operations for non-commuting operators ****)

SetAttributes[nc, {Flat, OneIdentity}]; (* Associativity *)

nc[c__] /; (Length[{c}] === 1) := c ;
nc[c___] /; (Length[{c}] === 0) := 1;

nc[a_, z_?NumericQ, b_] := z nc[a, b];
nc[z_?NumericQ, b_] := z b;
nc[a_, z_?NumericQ] := z a;

nc[a_, z_?NumericQ b_, c_] := z nc[a, b, c];
nc[z_?NumericQ b_, c_] := z nc[b, c];
nc[a_, z_?NumericQ b_] := z nc[a, b];

nc[a___, b1_ + b2_, c___] := nc[a, b1, c] + nc[a, b2, c];

(* Pauli rule *)
nc[op_[k__], op_[k__]] := 0 /; fermionQ[op];

ordering[op_] = EMPTY; (* Default ordering is "Empty band" *)
(* Override with ordering[op] = SEA or whatever *)

(* In the case of empty band, we always reorder so that the creation
   operators go to the left of the operator product *)
snegOrderedQ[op_, {x1_, x2_}] := OrderedQ[{x1, x2}] /;
  ordering[op] === EMPTY;

(* In the case of filled particle sea, we never reorder automatically, 
   unless the operators differ in last (n-2) indeces, because these 
   are then essentially different fermion species. This therefore
   leads to disentanglement of different flavors of operators!! *)
snegOrderedQ[op_, {x1:op_[i1_, k1_, l1__], x2:op_[i2_, k2_, l2___]}] :=
  If[{l1} =!= {l2}, OrderedQ[{x1, x2}], True] /;
  ordering[op] === SEA;


(* Canonical sorting:
- creation operators to the left, annihilation operators to the right
- sort by index
- sort by spin index
=> for trivial vacuum (no particles), this reduces to simple sorting 
   by all indexes!
*)

nc[ x1:op_[k1__], x2:op_[k2__] ] := -nc[x2, x1] + acmt[op, {k1}, {k2}] /; 
(fermionQ[op] && Not[snegOrderedQ[op, {x1, x2}]]);

overrideSEA[expr_] := 
  expr //. { HoldPattern[nc[x1:op_[k1__], x2:op_[k2__]]] :>
    -nc[x2, x1] + acmt[op, {k1}, {k2}] /; 
    (fermionQ[op] && Not[OrderedQ[{x1, x2}]])
  };

(* Canonical ordering for UNEQUAL operators *)

(* XXX: in some application it would be beneficial to disentangle
operators.. Would it make sense to write a package with no automatic
ordering at all? How many things this would break?  Would a custom
Simplify[] be required? *)

nc[ x1:op1_[i1_,j1___], x2:op2_[i2_,j2___] ] := -nc[x2, x1] /; 
(fermionQ[op1] && fermionQ[op2] && (op1 =!= op2) && 
Not[OrderedQ[{i1,i2}]]); (* Creation operators always moved to the left *)

(* If same type of operators, sort according to symbol *)
nc[ x1:op1_[i1_,j1___], x2:op2_[i2_,j2___] ] := -nc[x2, x1] /; 
(fermionQ[op1] && fermionQ[op2] && (op1 =!= op2) && (i1 === i2) &&
Not[OrderedQ[{x1,x2}]]);

(* acmt[] returns 1 if one operator is creation operator, and the other
operator is the annihilation operators, while all other indeces are equal. *)
acmt[op_, {i1_, j1___}, {i2_, j2___}] :=
    If[i1 =!= i2 && {j1} === {j2}, 1, 0];

(* NOTE: If an operator has indeces, an overloading acmt[] must be
   defined. *)
(* For example:
  acmt[c, {i1_, k1_, sp1_}, {i2_, k2_, sp2_}] :=
       If[i1 =!= i2, KroneckerDelta[k1, k2] KroneckerDelta[sp1, sp2], 0];
  defines a canonical commutator with Kronecker's deltas. 
*)

(* Commutators *)
komutator[a_, b_] := nc[a, b] - nc[b, a];
antikomutator[a_, b_] := nc[a, b] + nc[b, a];

(* (Manual) normal ordering *)
normalorder[a_] := a - vev[a];
normalorderwick[a_] := a - vevwick[a];

(* dd[] denotes a double dots surrounding an expression, i.e. a normal
ordered operator string. This is merely a placeholder. To explicitly
evaluate the normal ordered expression, use normalorder[] function. *)

sneglinearoperator[dd];

(* Associativity: XXX, check! *)
dd[a___, HoldPattern[dd[b__]], c___] := dd[a, b, c]; 

dd[] := 1;

(* Conjugation *)
conj[z_?NumericQ] := Conjugate[z];
conj[z_?NumericQ a___] := Conjugate[z] conj[a];
conj[a_ + b_] := conj[a] + conj[b];
conj[ HoldPattern[nc[a_, b_]] ] := conj[nc[b]] ~ nc ~ conj[a];

(* The following may not hold in general, but usually it does... *)
(* XXX: What to do with Majoranas? *)
conj[op_?operatorQ[i_, j___]] := op[If[i == CR, AN, CR], j];

(* Conjugation of sums *)
conj[sum[z_, it_List]] := sum[conj[z], it];

(* conj is Listable *)
conj[l_List] := conj /@ l;

(* conj threads over Rules *)
conj[Rule[a_, b_]] := Rule[conj[a], conj[b]];

(**** Symbolic sums ****)

(* No sum at all *)
sum[z_, {}] := z;

(* The indeces must be sorted for the simplifications to work correctly *)
sum[z_, it_List] := sum[z, Sort[it]] /; Not[OrderedQ[it]];

(* Returns True if expr contains none of symbols in list it *)
FreeOfIndexQ[expr_, it_List] := And @@ Map[FreeQ[expr, #]&, it];

(* Return True if expr can be factored out in front of the sum[] *)
sumFactorizableQ[expr_, it_List] := NumericQ[expr] && FreeOfIndexQ[expr, it];

(* It's always safe to extract a number *)
sum[z_ a_, it_List] := z sum[a, it] /; (NumberQ[z] && z != 1);

sum[z_, it_List] := z sum[1, it] /; (sumFactorizableQ[z, it] && z != 1);

nc[sum[a1__, it1_List], sum[a2__, it2_List]] := sum[nc[a1, a2], Join[it1,it2]];

(* Multiple sums *)
sum[sum[a_, it1_List], it2_List] := sum[a, Union[it1, it2]];

(* Thread the first argument *)
sumThread = {
sum[z_List, it_List] :> Map[sum[#, it]&, z]
}

sumExpand = { 
sum[a_ + b_, it_List] :> sum[a, it] + sum[b, it],

sum[z_ a_, it_List] :> z sum[a, it] /; (sumFactorizableQ[z, it]),

sum[a_, it_List] :> sum[Expand[a], it]
};

sumCollect = { 
z1_. sum[a1_, it1_List] + z2_. sum[a2_, it1_List]
  :> sum[z1 a1 + z2 a2, it1],

  HoldPattern[nc[a_, sum[b_, it_List], c_]] :> sum[nc[a, b, c], it]
};

sumSimplify = {
  sum[KroneckerDelta[n1_, n2_] a_., it_List] 
  :> sum[a //. {n1 -> n2}, Complement[it, {n1}]]
  /; MemberQ[it, n1],

  sum[KroneckerDelta[n1_, n2_] a_., it_List]
  :> sum[a //. {n2 -> n1}, Complement[it, {n2}]]
  /; MemberQ[it, n2],

  sum[KroneckerDelta[a__]^2 b_., it_List]
  :> sum[KroneckerDelta[a] b, it],

  sum[z_ a_, it_List] 
  :> z sum[a, it] /; (sumFactorizableQ[z, it]),

  HoldPattern[sum[a_ b_., {i___, n1_, j___}]] :>
  sum[a, {i, j}] sum[b, {n1}] /; FreeQ[a, n1] && FreeOfIndexQ[b, {i, j}],

  HoldPattern[sum[a_ (b_ + c_), x:{i___, n1_, j___}]] :>
  sum[a c, x] + sum[a, {i, j}] sum[b, {n1}] /; 
  FreeQ[a, n1] && FreeOfIndexQ[b, {i, j}]

(*  HoldPattern[sum[a_, {i___, n1_, j___}]] :>
  sum[a, {i, j}] sum[1, {n1}] /; FreeQ[a, n1] *)
};

(* Replace indexes with abstract indeces ndxfunc[i] with i=1,...,nrindeces *)
(* Be careful: this is only useful for fine-tuned applications to very
specific problems! *)
sumAbstractIndex[expr_, ndxfunc_] := Module[{ r },
  SetAttributes[ndxfunc, NumericFunction];
  r = sum[a_, it_List] :> Module[{rule, len},
        len = Length[it];
        rule = Table[it[[i]] -> ndxfunc[i], {i, len}];
        sum[a //. rule, it //. rule]
      ];
  expr /. r
];

sumNameIndex[expr_, ndxfunc_, li_List] := Module[ {len},
  len = Length[li];
  expr //. Table[ndxfunc[i] -> li[[i]], {i, len}]
];

(* Attempt to simplify expression involving symbolic summations. Heuristic
approach is used that usually works. TO DO: can we do better than this? *)

sumFullSimplify[expr_] := Module[{ex},
  ex = Expand[expr];
  ex = ex //. sumCollect;
  ex = Simplify[ex];
  ex = ex //. sumSimplify;
  ex = Simplify[ex];

  (* We must also expand within sum[] *)
  ex = ex //. sumExpand;
  ex = Simplify[ex];
  ex = ex //. sumSimplify;
  ex = Simplify[ex];

  (* and then collect again *)
  ex = ex //. sumCollect;
  ex = Simplify[ex]
];

(* Vacuum expectation value *)
sneglinearoperatorFirst[vev];
vev[op_?operatorQ[k__]] := 0; (* single operator *)

vev[ nc[a___, op_[AN, j___]] ] := 0 /; 
operatorQ[op] && ordering[op] === EMPTY; (* annihilates vacuum to the right *)

vev[ nc[op_[CR, j___], a___] ] := 0 /; 
operatorQ[op] && ordering[op] === EMPTY; (* annihilates vacuum to the left *)

vev[sum[ops_, it_List]] := sum[ vev[ops], it ];
vevwick[sum[ops_, it_List]] := sum[ vevwick[ops], it ];

(*** Vacuum expectation value for a Fermion sea ***)
(* The second index must be the wavenumber or an equivalent quantity *)

(* (1-n_k) |0> *)
vev[ nc[a_, op_[AN, k_, j___], op_[CR, k_, j___]] ] :=
  UnitStep[k] vev[nc[a]] /;
  operatorQ[op] && ordering[op] === SEA;
  
(* <0| (1-n_k) *)
vev[ nc[op_[AN, k_, j___], op_[CR, k_, j___], b_] ] :=
  UnitStep[k] vev[nc[b]] /;
  operatorQ[op] && ordering[op] === SEA;

(* n_k |0> *)
vev[ nc[a_, op_[CR, k_, j___], op_[AN, k_, j___]] ] :=
  UnitStep[-k] vev[nc[a]] /;
  operatorQ[op] && ordering[op] === SEA;

(* <0| n_k *)
vev[ nc[op_[CR, k_, j___], op_[AN, k_, j___], b_] ] :=
  UnitStep[-k] vev[nc[b]] /;
  operatorQ[op] && ordering[op] === SEA;
  
(* VEV of two operators with two different sets of ultimate indeces is zero.
Here we need to be extremely careful about the free indeces which can take
an arbitrary value. Note the four-argument version of the If[] statement! *)

vev[ nc[op_[i1_, k1_, j1___], op_[i2_, k2_, j2___]] ] := 0 /;
  operatorQ[op] &&
  ordering[op] === SEA && If[{j1} != {j2}, True, False, False];

(* Finally, an explicit procedure for bilinear forms in the case of SEA
ordering. This is then used for more general expression, for example in
conjunction with the Wick's theorem. *) 
(* XXX: How should I generalize this? *)

vev[ nc[op1_[i1_, k1_, j1___], op2_[i2_, k2_, j2___]] ] :=
  KroneckerDelta[op1, op2] KroneckerDelta[{j1}, {j2}] KroneckerDelta[k1, k2] *
  (If[i1 == CR && i2 == AN, UnitStep[-k1], 0]
  + If[i1 == AN && i2 == CR, UnitStep[k1], 0]) /; 
  operatorQ[op1] && operatorQ[op2] && 
  ordering[op1] === SEA && ordering[op2] === SEA;

(* Other cases need to be evaluated using Wick's theorem !! *)


(* Drops operators that annihilate vacuum. See also zeroonvac[]. *)
sneglinearoperatorFirst[onvac];
onvac[ nc[a_, op_[AN, j___]] ]  := 0 /; operatorQ[op];

(* Inner and outer products using nc *)
mdot[a_, b_] := Inner[nc, a, b];
inner[a_, b_] := Inner[nc, a, b];
outer[a_, b_] := Outer[nc, a, b];

(* Vector-matrix-vector product using mdot multiplication *)
VMV[v1_, m_, v2_] := (v1 ~ mdot ~ m) ~ mdot ~ v2;

(* Number operator *)
number[op_[j___], sigma_] := op[CR, j, sigma] ~ nc ~ op[AN, j, sigma];
number[op_[j___]] := number[op[j], UP] + number[op[j], DO];

(* Hubbard's local e-e repulsion operator *)
hubbard[op_[j___]]:= number[op[j], UP] ~ nc ~ number[op[j], DO];

(* The Hubbard model for operator op[], nn sites and parameters
   U, t and eps *)
hamiltonian[Hubbard, op_[j___], nn_, U_, t_, eps_] :=
  Sum[eps number[op[i, j]] + U hubbard[op[i, j]], {i, 1, nn}] +
  t Sum[hop[op[i, j], op[i+1, j]], {i, 1, nn-1}];

(* Inter-site charge-charge interaction *)
chargecharge[op1_[j1___], op2_[j2___]] :=
  number[op1[j1]] ~ nc ~ number[op2[j2]];

(* Spin operators (Pauli matrices) for S=1/2 fermions. Note that
they must be multiplied by 1/2 to obtain true spin operators. *)
PauliX = {{0, 1}, {1, 0}};
PauliY = {{0, -I}, {I, 0}};
PauliZ = {{1, 0}, {0, -1}};
PauliPlus = PauliX + I PauliY;
PauliMinus = PauliX - I PauliY;

(* Spin XYZ operators for some arbitrary generalized
   creation operator g, i.e. some interesting combination
   of one-site creation operators. 'g' must be a function
   of spin index 'sigma'. *)
spinxyzgen[g_] := Module[{cr, an},
  cr = {g /. sigma -> UP, g /. sigma -> DO};
  an = conj /@ cr;
  1/2 VMV[cr, #, an]& /@ {PauliX, PauliY, PauliZ}
];

(* Returns the x, y and z components of the spin 1/2 operator. *)
spinxyz[op_[j___]] := spinxyzgen[op[CR, j, sigma]];

(* Returns the x, y, and z components of the spin 1/2 operator, as well as
the and the S^2 operator *)
spinall[op_[j___]] := Module[{sxyz, ss},
  sxyz = spinxyz[op[j]];
  ss = sxyz ~ mdot ~ sxyz;
  Append[sxyz, ss]
];


(* Shorthand calls for generation of individual spin operators. *)
spinss[op_[j___]] := spinall[op[j]] [[4]];
spinx[op_[j___]] := spinall[op[j]] [[1]];
spiny[op_[j___]] := spinall[op[j]] [[2]];
spinz[op_[j___]] := spinall[op[j]] [[3]];

spinplus[op_[j___]] :=  Simplify[ spinx[op[j]] + I spiny[op[j]] ];
spinminus[op_[j___]] := Simplify[ spinx[op[j]] - I spiny[op[j]] ];

(* Scalar product of spin operators for op1 and op2 *)
spinspin[op1_[j1___], op2_[j2___]] := Module[{sp1, sp2},
  sp1 = spinxyz[op1[j1]];
  sp2 = spinxyz[op2[j2]];
  Simplify[ sp1 ~ inner ~ sp2 ] (* FIXED: 30.6.06 *)
];
  
(* Recall: S1 . S2 = S1^z S2^z + 1/2 (S1^+ S2^- + S1^- S2^+) *)
(* Note: spinspin[a,b] = spinspinz[a,b] + spinspinxy[a,b] -> no factor 1/2 
   in this definition! *)

(* Transverse part of the spin-spin interaction *)
spinspinpm[op1_[j1___],op2_[j2___]] := 
  spinplus[op1[j1]] ~ nc ~ spinminus[op2[j2]];

spinspinmp[op1_[j1___],op2_[j2___]] := 
  spinminus[op1[j1]] ~ nc ~ spinplus[op2[j2]];

spinspinxy[op1_[j1___],op2_[j2___]] := Module[{sp1, sp2},
  sp1 = spinspinpm[op1[j1], op2[j2]];
  sp2 = spinspinmp[op1[j1], op2[j2]];
  (sp1+sp2)/2
];

(* Longitudinal part of the spin-spin interaction *)
spinspinz[op1_[j1___],op2_[j2___]] := Module[{},
  spinz[op1[j1]] ~ nc ~ spinz[op2[j2]]
];
  
(* Given a list of operators 'ops', manyspin[] constructs the
components of the total spin, as well as the total spin squared operator. *)
manyspin[ops_List] := Module[{sxyz, ss},
  sxyz = Expand @ (Plus @@ Map[spinxyz, ops]);
  ss = Expand @ (sxyz ~ mdot ~ sxyz);
  Append[sxyz, ss]
];

(* Returns True if z is half-spin or entire spin *)
halfintegerQ[z_] := IntegerQ[2z];

(* Spin matrices for arbitrary spin *)
spinmatrixZ[S_] := 
    Table[If[sz1 == sz2, sz1, 0], {sz1, S, -S, -1}, {sz2, S, -S, -1}] /; 
      halfintegerQ[S];
spinmatrixP[S_] := 
    Table[If[sz1 == sz2 + 1, Sqrt[S(S + 1) - sz2(sz2 + 1)], 0], {sz1, 
          S, -S, -1}, {sz2, S, -S, -1}] /; halfintegerQ[S];
spinmatrixM[S_] := 
    Table[If[sz1 == sz2 - 1, Sqrt[S(S + 1) - sz2(sz2 - 1)], 0], {sz1, 
          S, -S, -1}, {sz2, S, -S, -1}] /; halfintegerQ[S];
spinmatrixX[S_] := 1/2(spinmatrixP[S] + spinmatrixM[S]) /; halfintegerQ[S];
spinmatrixY[S_] := 1/(2I)(spinmatrixP[S] - spinmatrixM[S]) /; halfintegerQ[S];


(* MISSING!!!! *)

(* Direct product of two basis *)
direct[l1_List, l2_List] := Flatten @ Outer[nc, l1, l2];

(* Nambu spinor, isospin operators *)
nambu[op_[j___], n_:0] := {op[CR, j, UP], (-1)^n op[AN, j, DO]};

isospin[op_[j___], n_:0] := Module[{nam, nbar},
  nam = nambu[op[j], n];
  nambar = conj /@ nam;
  1/2 VMV[nam, #, nambar]& /@ {PauliX, PauliY, PauliZ}
];

manyisospin[ops_List, ns_List] := Module[{len, ixyz, ii},
  len = Length[ops];
  ixyz = Expand @ (Plus @@ Table[isospin[ ops[[i]], ns[[i]] ], {i, len}]);
  ii = (ixyz ~ mdot ~ ixyz) // Expand;
  Append[ixyz, ii]
];

(* Basis states for a site *)
basis[op_[j___]]:= {1, op[CR,j,UP], op[CR,j,DO], nc[op[CR,j,DO], op[CR,j,UP]]};

(* Integer powers of operators *)
pow[op_, 1] := op;
pow[op_, i_Integer] := nc @@ Table[op, {i}];

(* Hopping operator *)
hop[op1_[j1___], op2_[j2___], sigma_] := 
  op1[CR, j1, sigma] ~ nc ~ op2[AN, j2, sigma] +
  op2[CR, j2, sigma] ~ nc ~ op1[AN, j1, sigma];

hop[op1_[j1___], op2_[j2___]] :=
  hop[op1[j1], op2[j2], UP] + hop[op1[j1], op2[j2], DO];

(* Hopping with spin-flip *)
spinfliphop[op1_[j1___], op2_[j2___]] := Sum[
  op1[CR, j1, sigma] ~ nc ~ op2[AN, j2, 1-sigma] +
  op2[CR, j2, sigma] ~ nc ~ op1[AN, j1, 1-sigma],
{sigma, DO, UP} ];

(* Hopping with hole operators: 
c^\dag_{k,\sigma} f_\sigma} + f^\dag_\sigma c_{k\sigma}
\to 
h_{k,-\sigma} f_\sigma + f^\dag_\sigma h^\dag_{k,-\sigma}
*)

holehop[oph_[jh___], opf_[jf___], sigma_] := 
oph[AN, jh, sigma] ~ nc ~ opf[AN, jf, 1-sigma] +
opf[CR, jf, sigma] ~ nc ~ oph[CR, jh, 1-sigma];

holehop[oph_[jh___], opf_[jf___]] := 
  Sum[holehop[oph[jh], opf[jf], {sigma, DO, UP}]];

(* 2-electron hopping operator *)
twohop[op1_[j1___], op2_[j2___]] :=
  nc[ op1[CR, j1, DO], op1[CR, j1, UP], op2[AN, j2, DO], op2[AN, j2, UP] ] +
  nc[ op2[CR, j2, DO], op2[CR, j2, UP], op1[AN, j1, DO], op1[AN, j1, UP] ];  



(****************** MISC ********************)

sneglinearoperator[invertspin];

invertspin[] := 1;
invertspinindex[j_] := (1-j); (* This works for S=1/2 only! *)

(* Inverts the spin of all operators (spin reflection) *)
invertspin[op_[i_, j_, k___]] := 
  op[i, invertspinindex[j], k] /; operatorQ[op];

invertspin[HoldPattern[ nc[a__] ]] := nc @@ Map[invertspin, {a}];


(*********** SIMPLIFICATION RULES ***********)

ruleCollect = { z1_ a_nc + z2_ a_nc :> (z1+z2) a };

(************** AP **************)


(* BASIS holds a (global) list of creation operators for all possible
   single-particle states. It is constructed using a call to makebasis[],
   see below. Creation operators for all many-particle states in the (Q,S)
   basis are constructed by combining these operators using suitable
   Clebsch-Gordan coefficients. *)

BASIS = {};

(* Construct basis for common enlisted operators.
   Example: makebasis[{a[],f[1]}] 
   Expansion over both spin orientations is automatically done. *)

mbfunc = {# /. op_[j___] -> op[CR, j, UP], 
   # /. op_[j___] -> op[CR, j, DO]} &;
makebasis[{l__}] := makebasis[l];
makebasis[l__] := BASIS = (mbfunc /@ {l}) // Flatten;

(* Is the operator in the basis? Disregards CR vs. AN type. *)
goodopQ[op_[i_, j___]] := Count[BASIS, op[CR, j]] == 1;

(* Mapping of operators to positions in state vectors. An index
   is only returned if the operator actually exists in the basis,
   otherwise the function call is unevaluated. *)
op2ndx[op_[i_, j___]] := Position[BASIS, op[CR, j]] [[1, 1]] /; 
  goodopQ[op[i, j]];

ndx2op[n_Integer] := BASIS[[n]];

(* Vacuum vector: vc[0, 0, ..., 0] *)
vacuum[] := vc @@ Table[0, {Length[BASIS]}];

(*** Applies an operator to a vector in occupation number representation ***)
 
sneglinearoperator[ap];

ap[v_vc] := v;

ap[HoldPattern[x:nc[___]], v_vc] := 
  Fold[ ap[#2, #1]&, v, Reverse[ List @@ x] ];

ap[op_[i_, j___], v_vc] := Module[{ndx, sign, ok},
  ndx = op2ndx[op[i, j]];
  sign = If[EvenQ[Length[Cases[Take[v, ndx-1], 1]]], 1, -1];
  ok = If[v[[ndx]] == i, 1, 0];
  ok sign MapAt[(1-#)&, v, ndx] 
] /; goodopQ[op[i, j]];

(* Apply list of operators *)
ap[{l__}, b__] := ap[#, b]& /@ {l};

(* Transform a vector in number representation to a 
  corresponding creation operator *)

sneglinearoperator[vc2ops];

vc2ops[vc[l__]] := Module[{a},
  a = Inner[Times, BASIS, {l}, ncx];
  a = Select[a, (# =!= 0)&];
  a /. ncx->nc
];

(* Apply vc2ops to a List *)
vc2opsmap[l_List] := vc2ops /@ l;


(* It is very important that the integers in vc[] are not converted to real
values by application of N[] to some state. *) 
SetAttributes[vc, NHoldAll];

(* Scalar product of expressions with number representation vectors *)
sneglinearoperator[scalarproduct];
scalarproduct[a_vc, b_vc] := If[a === b, 1, 0];

scalarproductvc[a___] = scalarproduct[a]; (* an alias *)

(* This is essentially a bracket of two operator expressions *)
scalarproductop[a_, b_] := vev[conj[a] ~ nc ~ b];

(* Helper function: this is a bracket <cr|op|cr> *)
expvop[op_, cr_] := Module[{ levo, desno },
  levo = conj[cr] // Expand;
  desno = op ~ nc ~ cr // Expand;
(* desno = onvac @ desno; *)
  vev[levo ~ nc ~ desno // Expand]
];

(* Construct a matrix representation of operator 'a' in basis 'l'.
   Note: 'l' must be expressed in occupation number representation. *)
matrixrepresentationvc[a_, l_List] := Module[{bl},
  bl = Map[ap[a, #]&, l];
  Outer[scalarproduct, l, bl]
];

(* Construct a matrix representation of operator 'a' between two
   different subspaces 'l1' and 'l2'. States 'l1' are bras,
   while states 'l2' are kets. *)
matrixrepresentationvc[a_, l1_List, l2_List] := Module[{bl},
  bl = Map[ap[a, #]&, l2];
  Outer[scalarproduct, l1, bl]
];

(* As above, but for operator expressions *)
matrixrepresentationop[a_, l1_List, l2_List] := Module[{bl},
  bl = Map[nc[a, #]&, l2];
  Outer[scalarproductop, l1, bl]
];



(************* (Q,Sz) BASIS **************)

(* Make a basis of good quantum numbers Q and S_z. This is trivially done in
the occupation number representation vc[...]. *) 

qszbasisvc[l_List] := Module[{n = Length[l], ndxs},
  ndx2digs[ndx_] := IntegerDigits[ndx, 2, n];

  qunum[up_, do_] := Module[{iup, ido},
    iup = Plus @@ ndx2digs[up];
    ido = Plus @@ ndx2digs[do];
    {iup+ido-n , 1/2 (iup-ido)}
  ]; (* Q ans S_z *)

  ndx2vc[up_, do_] := vc @@ interleave[ndx2digs[up], ndx2digs[do]];
  
  ndxs = Table[ {qunum[up, do], ndx2vc[up, do]}, {up, 0, 2^n-1}, {do, 0, 2^n-1}];
  ndxs = Flatten[ndxs, 1];
  mergebasis[ndxs]
];

(* IMPORTANT: It is assumed that makebasis[l] had been called with the same l. *)

qszbasis[l_List] := Module[{bzvc},
  bzvc = qszbasisvc[l];
  Map[ {#[[1]], Map[vc2ops, #[[2]]]}&, bzvc]
];

(* Interleave lists *)
interleave[l__] := Flatten[ Transpose[{l}], 1];

(************* (Q,S) BASIS ***************)

(* Make a basis of good quantum numbers Q and S *)
(* Convention: always use the highest possible Sz, i.e. Sz == S ! *)

(* Change 0<->1 in the last index of an operator. This is just a faster
version of the more general invertspin[] function. *)
spinflip[op_[i_, j___, sigma_]] := op[i, j, 1-sigma];

(* We start with an empty list of (Q,S) states and the full list
   of all single-particle operators. *)
qsbasis[l_List] := qsbasis[{}, l];

(* Last step: when no other operators need to be added, we're done! *)
qsbasis[m_List, {}] := m;

(* First step: construct the three possible (Q,S) states from one
   single-particle creation operator. *)
qsbasis[{}, l_List] := Module[{op1, op2},
  op1 = l[[1]] /. op_[j___] -> op[CR, j, UP];
  op2 = spinflip[op1];
  qsbasis[ 
    {
      { {-1, 0}, {1} },
      { {0, 1/2}, {op1} },
      { {1, 0}, {nc[op2, op1]} }
    }, Drop[l, 1]
  ]
];

(* Intermediate steps: add one single-particle operator from list 'l'
   and construct suitable combination of operators with a well-defined
   (Q,S) quantum numbers. This step is reiterated until completion. *)
qsbasis[m_List, l_List] := Module[{mnew},
  mnew = keymerge[ ADD[m, l[[1]]] ];
  qsbasis[mnew, Drop[l, 1]]
];

(* Returns 0 for operator expressions which annihilate vacuum *)
sneglinearoperatorFirst[zeroonvac];
zeroonvac[op_?operatorQ[i_, j___]] := If[i === AN, 0, op[i, j]];
zeroonvac[ x:nc[a_, op_?operatorQ[i_, j___]] ] :=
  If[i === AN, 0, x];

(* XXX: V cem se zeroonvac razlikuje od onvac?? onvac[] ostane okoli
izrazov. Ga kje uporabljam???*)

(* NEW: let's apply zeroonvac automatically if so requested. *)
(* This might turn out to be a neat idea. *)

nc[a_, op_[AN, j___], VACUUM] := 0;
nc[conj[VACUUM], op_[CR, j___], a_] := 0;

(* Vacuum state must be normalized to 1! *)
vev[ HoldPattern[nc[conj[VACUUM], VACUUM]] ] = 1;

(* Smart spindown: determines which operators appear in the expressions,
builds a corresponding spin down operator and applies it to the expression. *)
SPINDOWN[a_] := nc[Plus @@ Map[spinminus, Union[Cases[a, 
op_?operatorQ[i_, j___, sigma_] -> op[j], {0, Infinity} ]]], a];

(* Calculate a 'norm' for an operator expression, i.e. sqrt(<expr|expr>). *)

normop[expr_] := Sqrt @ vev[ scalarproductop[expr, expr] ];
normvc[expr_] := Sqrt @ vev[ scalarproductvc[expr, expr] ];

(* applytocr[] applys an operator to an expression, drops parts which
annihilate the vacuum and "normalizes" the reminder *)
applytocr[op_, expr_] := Module[{b},
  b = zeroonvac[nc[op, expr]];
  b / normop[b]
];

(* Smarter spindown: drops parts which annihilate vacuum and "normalizes"
the remainder *)
spindown[a_] := Module[{b},
  b = zeroonvac[SPINDOWN[a]];
  b / normop[b]
];

(* Get sublists with a given key *)
keyselect[l_, key_] := Select[l, #[[1]] == key &];

(* Merge sublists with the same first element (key) *)
(* Below there is another implementation mergebasis[] which does the same! *)

keymerge[l_] := Module[{keys, sublists},
  keys = Union[ l[[All,1]] ];
  sublists = Map[keyselect[l, #]&, keys];
  Table[{keys[[i]], sublists[[i,All,2]] // Flatten}, {i, Length[keys]}]
];

(* Add one (doublet) operator to a (Q,S) basis. This is used
   by qsbasis[] to generate the entire list of basis states. *)
ADD[l_List, x:op_[j___]] :=
  Map[ADDQS[#, x]&, l] ~ Flatten ~ 1;

ADDQS[l_List, x:op_[j___]] :=
  Map[ADDONE[#, x]&, Map[{l[[1]],#}&, l[[2]] ] ] ~ Flatten ~ 1;

ADDONE[{{q_, s_}, a_}, op_[j___]] := Module[{op1, op2, res},
  op1 = op[CR, j, UP];
  op2 = spinflip[op1];
  res = {
    { {q-1, s}, {a} },
    { {q+1, s}, {nc[op2, op1, a]} },
    { {q, s+1/2}, {nc[op1, a]} }
  };
  If[s>0, Append[res, { 
    {q, s-1/2}, { Sqrt[2s/(2s+1)] nc[op2, a] - 
    Sqrt[1/(2s+1)] nc[op1, spindown[a] ] } 
  }], res] (* cf. Krishna-Murthy et al. *)
];

(**** Projection operators ****)

(* Indexing is the same as in T. Rejec's operachain.cpp code for
   evaluation of operator expressions *)

(* Short-cuts *)
ClearAttributes[{PROJ0, PROJUP, PROJDO, PROJ2, PROJ1, PROJ02}, Protected];

PROJ0 = 1;   
PROJUP = 2;
PROJDO = 3;
PROJ2 = 4;
PROJ1 = 5;
PROJ02 = 6;

SetAttributes[{PROJ0, PROJUP, PROJDO, PROJ2, PROJ1, PROJ02}, Protected];

(* Subspace with 0 particles *)
projector[op_[j___], 1] := nc[ 1-number[op[j], UP], 1-number[op[j], DO] ];
projector0[op_[j___]] := projector[op[j], PROJ0];

(* Subspace with spin up particle *)
projector[op_[j___], 2] := nc[   number[op[j], UP], 1-number[op[j], DO] ];
projectorUP[op_[j___]] := projector[op[j], PROJUP];

(* Subspace with spin down particle *)
projector[op_[j___], 3] := nc[   number[op[j], DO], 1-number[op[j], UP] ];
projectorDO[op_[j___]] := projector[op[j], PROJDO];

(* Subspace with two particles *)
projector[op_[j___], 4] := nc[   number[op[j], UP],   number[op[j], DO] ];
projector2[op_[j___]] := projector[op[j], PROJ2];

(* Subspace with one particle of either spin *)
projector[op_[j___], 5] := projector[op[j], 2] + projector[op[j], 3];
projector1[op_[j___]] := projector[op[j], PROJ1];

(* Subspace with occupancy 0 or 2 *)
projector[op_[j___], PROJ02] := 
  projector[op[j], PROJ0] + projector[op[j], PROJ2];
projector02[op_[j___]] := projector[op[j], PROJ02];

(**** Operations on basis sets ****)
bzvc2bzop[bvc_] := Map[{#[[1]], vc2opsmap[#[[2]]]}&, bvc];
bzop2bzvc[bz_, vak_] := Map[{#[[1]], ap[#[[2]], vak]}&, bz];

(* Decompose a vector in a given basis *)
decomposeop[vec_, basis_] := Map[scalarproductop[#, vec]&, basis];
decomposevc[vec_, basis_] := Map[scalarproduct[#, vec]&, basis];


(**** Orthogonalization ****)
Needs["LinearAlgebra`Orthogonalization`"];

orthogop[vecs_, basis_] := Module[{m, l},
  m = Map[ decomposeop[#, basis]&, vecs ];
  l = Householder[m];
  l = Select[l, Norm[#] != 0 &]; (* Drop zero vectors *)
  Expand @ Map[(# . basis)&, l]
];

orthogvc[vecs_, basis_] := Module[{m, l},
  m = Map[ decomposevc[#, basis]&, vecs ];
  l = Householder[m];
  l = Select[l, Norm[#] != 0 &]; (* Drop zero vectors *)
  Expand @ Map[(# . basis)&, l]
];

(*** Projecting out a state ***)

(* Project out 'state' from a space of 'vecs'. Is is assumed that
   both 'state' and 'vecs' are normalized to 1.
   The resulting vectors are reorthogonalized. *)
remove1state[vecs_, state_] := Module[{vecs2},
  If[vecs != {},
      vecs2 = Map[# - scalarproduct[state, #] state &, vecs];
      orthogvc[vecs2, vecs] // Expand,
      {}
  ]
];

(* Project out 'states' from a space of 'vecs'. *)
removestates[vecs_, states_] := Module[{},
  Fold[remove1state, vecs, states]
];

(** Transformation from basis with well defined (Q,S) to a basis
   with well defined (I,S) quantum numbers **)
transformQStoIS[baza_] := transformQStoIS[Reverse[baza], {}];
transformQStoIS[{}, novabaza_] := Select[novabaza, (#[[2]] != {}) &];
transformQStoIS[baza_, novabaza_] := Module[{nova2, ISv, ostali},
  {ISv, ostali} = transform1[baza];
  nova2 = Append[novabaza, ISv];
  transformQStoIS[ostali, nova2]
];

(* baza is a basis List in the format as returned by function qsbasis[] *)
(* Tminus is the isospin lowering operator that we apply to the vectors in
   the subspace that we're subtracting from other subspaces *)
(* TO DO: Check if Simplify[] can be replaced by Expand[]. *)
transform1[baza_] := 
Module[{prvi, ostali, subspace, states, statesplus, pos, ISvec, i},
  prvi = First[baza];  (* The subspace that we are subtracting *)
  ostali = Rest[baza]; (* The subspaces that we are subtracting from *)
  subspace = prvi[[1]];(* (Q, S) indeces *)
  states = prvi[[2]];  (* Vectors *)

  (* Vectors, to which we are consecutivly applying J + *)
  statesplus = states;
      
  For[i = 1, True, i++,
    statesplus = Map[ap[Tminus, #] &, statesplus];
    statesplus = Expand @ statesplus; (* WAS: Simplify *)
    statesplus = DeleteCases[statesplus, 0]; (* Drop nul vectors *)
    If[statesplus == {},
      Break[]; (* We're done *)
    ];

    (* Normalize to 1 *)
    statesplus = Map[#/normvc[#] &, statesplus];
    statesplus = Expand @ statesplus; (* WAS: Simplify *)
        
    (* Find relevant subspaces from which we can project out *)
    pos = Position[ostali, x_ /; x[[1]] == subspace - {2i, 0}, {1}, 
                   Heads -> False];
        
    If[pos == {},
      Break[],
      (* Performing the projecting *)
      ostali = MapAt[{#[[1]], removestates[#[[2]], statesplus]} &, 
              ostali, pos];
    ];
  ];
      
  (* Return {(I, S) pair, vectors} and the matrix of remaining vectors *)
  ISvec = {{1/2, 1}subspace, states};
  {ISvec, ostali}
];

(*** BRAKET ***)

(* Dirac's bra-ket notation for arbitrary states. This is useful to represent
non-fermionic degrees of freedom, such as bosons, spins, rotators, etc. *)

(* Convention: always commute bras and kets to the right of the
fermionic operators *)

nc[ x1:bra[b___], x2:op_[i___] ] := nc[x2, x1] /; operatorQ[op];
nc[ x1:ket[k___], x2:op_[i___] ] := nc[x2, x1] /; operatorQ[op];

(* Returns a list with True at non-Null positions *)
nonnullpattern[l_List] := (# =!= Null)& /@ l;

(* Returns True if there are no conflicting non-Null elements in two lists. This
is used in concatenating kets and bras. *)
compatiblepattern[l1_List, l2_List] := (Length[l1] == Length[l2]) && (Nor @@ 
  Thread[And[nonnullpattern[l1], nonnullpattern[l2]]]);

(* Returns True if non-Null elements come in pairs. This is used in calculating
scalar product, i.e. expectation value. *)
pairpattern[l1_List, l2_List] := (Length[l1] == Length[l2]) && !(Or @@
  Thread[Xor[nonnullpattern[l1], nonnullpattern[l2]]]);

(* Return position that do not contain the Null element *)
nonnullpos[l_List] := Position[l, _?(# =!= Null &), {1, Infinity}, Heads->False];

(* Default rule for expectation values: Kronecker delta over all non-Null indeces *)
braketrule[bra[b___], ket[k___]] := Module[{pos, l1, l2},
  pos = nonnullpos[{b}];
  l1 = Extract[{b}, pos];
  l2 = Extract[{k}, pos];
  Inner[KroneckerDelta, l1, l2, Times] 
] /; pairpattern[{b}, {k}];

(* When bra and ket are juxtaposed, we can evaluate the scalar product. We
only do this when the bra and ket are compatible in the sense that both have
the same pair pattern: quantum numbers at equivalent position are defined
either in both bra and ket or in neither. *)

nc[x1:bra[b___], x2:ket[k___]] := braketrule[x1, x2] /; 
  pairpattern[{b}, {k}];

(* bkcombine[]: combine two lists in such a manner that Null elements
are treated as placeholders that are filled only if the corresponding
element in the other list is non-Null. Used in concatenation of bras
and kets.*) 
(* TO DO: This could be make to work even for multi-dimensional
lists. *)

bkcombine::incompat = "Incompatible pattern of non-Null elements";
bkcombine[l1_List, l2_List] := Module[{pos},
  pos = nonnullpos[l2];
  If[pos =!= {}, ReplacePart[l1, l2, pos, pos], l1]
] /; compatiblepattern[l1, l2];
bkcombine[l1_List, l2_List] := (Message[bkcombine::incompat]; error[bkcombine, l1,l2]) /;
  !compatiblepattern[l1, l2];

(* Direct product: concatenates all indeces, if possible! *)
nc[x_[i1___], x_[i2___]] := x @@ bkcombine[{i1}, {i2}] /;
  (x === bra || x === ket) && compatiblepattern[{i1}, {i2}];

(* Conjugation *)
conj[ket[k___]] := bra[k];
conj[bra[b___]] := ket[b];

(* Spin states and spin operators *)
spinket[s_?halfintegerQ] := ket /@ Range[s, -s, -1];
spinbra[s_?halfintegerQ] := bra /@ Range[s, -s, -1];

spinketbraZ[s_?halfintegerQ] := VMV[spinket[s], spinmatrixZ[s], spinbra[s]];
spinketbraP[s_?halfintegerQ] := VMV[spinket[s], spinmatrixP[s], spinbra[s]];
spinketbraM[s_?halfintegerQ] := VMV[spinket[s], spinmatrixM[s], spinbra[s]];
spinketbraX[s_?halfintegerQ] := VMV[spinket[s], spinmatrixX[s], spinbra[s]];
spinketbraY[s_?halfintegerQ] := VMV[spinket[s], spinmatrixY[s], spinbra[s]];

(* Generates a basis for spin operators. The key is the value of
   the Z-component of spin, S_z. *)
spinbasis[s_?halfintegerQ] := {{#}, {ket[#]}}& /@ Range[s, -s, -1];

(* Apply rule to the basis states (vektor form) in baza *)
transformfunc[baza_, rule_] := Module[{vecs},
  vecs = Map[Simplify[ ap[rule, #] ]&, baza];
  orthogvc[vecs, baza]
];

(* Apply rule to an entire basis in the form of a list of {{Q. numbers}, 
{basis states}}... *)
transformbasis[bvc_List, rule_] := Module[{bvcnew},
  bvcnew = Map[{#[[1]], transformfunc[ #[[2]], rule ]}&, bvc];
  bvcnew = Select[bvcnew, (Length[ #[[2]] ] != 0) &]
];

(* Apply several rules to each basis element. *)
(* XXX: Untested! *)
transformbasismulti[bvc_List, rules_List] := Module[{bvcnew},
  bvcnew = Map[transformbasis[bvc, #]&, rules];
  bvcnew = Flatten[bvcnew, 1];
  mergebasis[bvcnew]
];

(* Merge the basis: gather basis states with the same (Q,S) pair. *)
mergebasis[bvc_List] := Module[{qn},
  qn = Union @ bvc[[All,1]]; (* Get all different quantum numbers. *)
  Map[{#, collectvalues[#, bvc]}&, qn]
];

(* Get all basis vectors with the same quantum numbers (key). *)
collectvalues[key_, bvc_List] := Flatten @ (Select[bvc, (#[[1]] === key)&] [[All,2]]);

(* Drop subspaces with no states from a given basis bvc *)
dropemptysubspaces[bvc_List] := Select[bvc, #[[2]] =!= {} &];

(* Apply a function to an entire basis. *)
applybasis[bvc_List, fn_] := Map[{#[[1]], Map[fn, #[[2]] ]}&, bvc];

(* Note: spin basis in S representation can be conveniently
constructed using qsbasis[] which includes a pseudofermion that
corresponds to the spin, and then projecting out zero and double
occupancy states, and finally replacing the remaining single particle
pseudofermion operators with spin kets. One should be careful about
the anticommutation operators, though. The pseudospin operator must be
moved to the rightmost position of the operator string by overriding
the conventional (anti)commutation rules. Of course, this only works
for spin 1/2. *)

QSpseudospin[bvc_List, x:op_?operatorQ[]] := Module[{vecs},
  vecs = transformbasis[bvc, projector[x, PROJ1]];
  vecs = applybasis[vecs, vc2ops];
  vecs = vecs /. op[j___] :> hold[op[j]];
  vecs = vecs //. nc[x1:hold[op[j1___]], x2:b_?operatorQ[j2___]] :>
    -nc[x2, x1];
  vecs = vecs /. hold[op[i_, sp_, j___]] :> ket[If[sp == UP, 1/2, -1/2]]
];

(* Application of an operator to a occupation number representation vector *)
ap[x:ket[___], vc[]] := vc[x];
ap[x:ket[___], vc[j__]] := vc[j, x] /; Head[Last[{j}]] =!= ket;
ap[ket[i___], vc[j___, ket[k___]]] := vc[j, ket @@ bkcombine[{i}, {k}]] /; 
  compatiblepattern[{i}, {k}];
ap[a___, x1:bra[i___], vc[j___, x2:ket[k___]]] := ap[a, vc[j]] braketrule[x1, x2] /;
  pairpattern[{i}, {k}];

(* Direct product of (Q,S) basis and a (single) phonon basis with up
to Nph phonons *)

transformQStoPH[bvc_List, Nph_Integer] := Module[{vecs, fn, phbasis},
  phbasis = Table[ket[i], {i, 0, Nph}];
  fn[x_] = ap[x, #]&;
  vecs = Map[applybasis[bvc, fn[#]]&, phbasis];
  mergebasis @ Flatten[vecs, 1]
];

(* Phonon hamiltonian construction functions *)
phononnumber[Nph_Integer] := Sum[i nc[ket[i], bra[i]], {i, 0, Nph}];
phononplus[Nph_Integer] := Sum[ Sqrt[i+1] nc[ket[i+1], bra[i]], {i, 0, Nph-1}];
phononminus[Nph_Integer] := Sum[ Sqrt[i] nc[ket[i-1], bra[i]], {i, 1, Nph}];


(*** LEFT/RIGHT SYMMETRY SUPPORT ***)

(* Given l, the list of operators, lrmap[] returns the transformation
   rules that perform the left/right transformation. *)

lrmap[l_List] := Module[{l1},
  l1 = Flatten @ (mbfunc /@ l);
  l2 = Flatten @ (mbfunc /@ Reverse[l]);
  Thread[l1 -> l2]
];

(* Perform an operation o on each subspace vectors *)

mapbasis::compat = "Incompatible subspaces";
mapbasis[bz1_, bz2_, o_:(1/2 Plus[##]&)] := Module[
  {subs1, vecs1, subs2, vecs2, vecs},
  subs1 = bz1[[All, 1]];
  vecs1 = bz1[[All, 2]];
  subs2 = bz2[[All, 1]];
  vecs2 = bz2[[All, 2]];
  If[subs1 =!= subs2, 
    Message[mapbasis::compat]; 
    Return[];
  ];
  
  vecs = Table[o[vecs1[[i]], vecs2[[i]]], {i, Length[vecs1]}];

  Transpose[{subs1, vecs}]
];

(* Orthogonalize bz1 with respect to original basis bz *)
orthogbasisop[bz1_, bz_] := mapbasis[bz1, bz, orthogop];
orthogbasisvc[bvc1_, bvc_] := mapbasis[bvc1, bvc, orthogvc];

(* Append a quantum number 'qn' to the good quantum numbers of each subspace
in basis bz *)
appendquantumnumber[bz_, qn_] := Map[{ Append[#[[1]], qn], #[[2]] }&, bz];

(* Putting it all together: transform a given basis to an equivalent basis
which takes into account the l/r symmetry, i.e. parity. l is the list of 
consecutive operators that are fed to lrmap[] function. *)

transformtoLR[bz_, l_] := Module[{presl, bz2, bzsim, bzasim},
  presl = lrmap[l];
  bz2 = bz /. presl;
  bzsim  = mapbasis[bz, bz2, 1/2 Plus[##]& ];
  bzasim = mapbasis[bz, bz2, 1/2 Subtract[##]& ]; 
  bzsim  = dropemptysubspaces @ orthogbasisop[bzsim, bz];
  bzasim = dropemptysubspaces @ orthogbasisop[bzasim, bz];
  bzsim  = appendquantumnumber[bzsim, 1];
  bzasim = appendquantumnumber[bzasim, -1];
  Sort @ Join[bzsim, bzasim]
];

transformtoLRvc[bz_, l_, vak_] := Module[
  {bvc, presl, bz2, bzsim, bzasim, bvcsim, bvcasim},
  bvc = bzop2bzvc[bz, vak];
  presl = lrmap[l];
  bz2 = bz /. presl;
  bzsim  = mapbasis[bz, bz2, 1/2 Plus[##]& ];
  bzasim = mapbasis[bz, bz2, 1/2 Subtract[##]& ]; 
  bvcsim  = bzop2bzvc[bzsim, vak];
  bvcasim = bzop2bzvc[bzasim, vak];
  bvcsim  = dropemptysubspaces @ orthogbasisvc[bvcsim,  bvc];
  bvcasim = dropemptysubspaces @ orthogbasisvc[bvcasim, bvc];
  bvcsim  = appendquantumnumber[bvcsim, 1];
  bvcasim = appendquantumnumber[bvcasim, -1];
  Sort @ Join[bvcsim, bvcasim]
];
  

(** Transformations of basis sets **)

(* Produce a vector with 1 on i-th site in a length-n vector *)
enain[i_, n_] := Table[If[j == i, 1, 0], {j, n}];

(* Given a set of operators "oldbasis", a set of operators "newbasis" and
a set of operator expressions for newbasis, produce the transformation
rules from the oldbasis to the new basis! *)
snegold2newrules[oldbasis_, newbasis_, rules_] := Module[{n, ob, mat, x},
  n = Length[oldbasis];
  If[n != Length[newbasis], 
    Print["Mismatching lengths."];
    Return[];
  ];
  ob = Table[oldbasis[[i]] -> enain[i, n], {i, Length[oldbasis]}];
  mat = rules /. ob;
  x = Norm[mat . Transpose[Conjugate[mat]] - IdentityMatrix[n]];
  If[x != 0,
    Print["The transformation rule is not unitary."];
    Return[];
  ];
  mat = Inverse[mat];
  Thread[Rule[oldbasis, mat . newbasis]]
];


(*** WICK'S THEOREM ***)

(* Produce all possible pairs with elements taken from two lists *)
allpairs[m_List, n_List] := Flatten[Outer[List, m, n], 1];

isCR[op_[t_, j___]] := If[t === CR, 1, 0];
isAN[op_[t_, j___]] := If[t === AN, 1, 0];

(* Return a list of positions of CR/AN opperators *)
indecesCR[HoldPattern[nc[l__]]] := Flatten @ Position[Map[isCR, {l}], 1];
indecesAN[HoldPattern[nc[l__]]] := Flatten @ Position[Map[isAN, {l}], 1];


(* Return the maximal number of pairs that can be contracted. *)
maxcran[HoldPattern[x : nc[l__]]] := Module[{nrcr, nran},
  nrcr = Length @ indecesCR[x];
  nran = Length @ indecesAN[x];
  Min[nrcr, nran]
];
  
(* Returns True if there are no overlapping elements in the list of pairs *)  
disjointQ[{l__}] := 
  Length[{l}] == 1 || Length[Union[l]] == (Plus @@ Map[Length, {l}]);

(* Conflicting lists of pairs are zeroed *)
zeroconflict[z_List] := Map[If[disjointQ[#], #, 0] &, z];
  
(* Drop 0 elements from a List *)
dropzeros[l_List] := Delete[l, Position[l, 0, {1}]];

(* Drop lists of pairs that have overlapping indecs *)
dropconflicts[l_List] := dropzeros @ zeroconflict[l];

(* Generate lists of nr pairs to be contracted *)
contractedpairs[nr_, l_List] := dropconflicts @ KSubsets[l, nr];

(* Contraction of two operators according to the standard definition:
substract the normal ordered expression. *)
contraction[x1 : op1_[t1_, i1__], x2 : op2_[t2_, i2__]] :=
    nc[x1, x2] - normalorder[x1 ~ nc ~ x2];

(* Contract one pair.  ctr is an order index that replaces the contracted
operators in the operator string, factor is the prefactor of the string and
ops is the string that we are currently applying the contraction to. *)

contractone[{ctr_, factor_, ops_List}, pair_List] := 
  Module[{l = ops, i1, i2, c},
    i1 = pair[[1]];
    i2 = pair[[2]];
    c = contraction[l[[i1]], l[[i2]]];
    l[[i1]] = ctr;
    l[[i2]] = ctr;
    {ctr + 1, c*factor, l}
  ];

(* Return the remaining operator string. The integer numbered indexed
placeholders are commuted to the beginning of the list and ordered: this
gives the correct sign of the remaining operator string. Sign is flipped
whenever two neighboring integers are transposed, and when an integer and a
neighboring operator are transposed. Note: List2 is a non-Listable List. *)

remainder[{ops__}] := Module[{ops2}, 
  ops2 = List2[ops] //. {List2[a___, nr1_Integer, nr2_Integer, b___] :> 
    -List2[a, nr2, nr1, b] /; nr2 < nr1,
    List2[a___, x : c_[i__], nr_Integer, b___] :> 
    -List2[a, nr, x, b] /; operatorQ[c]
  };
  ops2 = ops2 /. {x : List2[i__] :> DeleteCases[x, _Integer, {1}]};
  ops2 = ops2 /. {List2[i___] :> dd[nc[i]]}
];
  
(* Contruct "pairs" in the operator string "ops" *)  

contract[ops_List, pairs_List] := Module[{ctr, factor, indexes},
  {ctr, factor, indexes} = Fold[contractone, {1, 1, ops}, pairs];
  remainder[indexes] * factor
];

(* Do the Wick! *)

wickorder[0, HoldPattern[x : nc[l__]]] := dd[x];

wickorder[nr_ /; nr > 0, HoldPattern[x : nc[l__]]] := Module[{ic, ia, p, cp},
  ic = indecesCR[x];
  ia = indecesAN[x];
  p = allpairs[ic, ia];
  cp = contractedpairs[nr, p];
  Plus @@ Map[contract[{l}, #] &, cp]
];

sneglinearoperator[wick];

(* Write an operator string using normal ordered strings and contractions *)
wick[HoldPattern[x : nc[l__]]] := Module[{m},
  m = maxcran[x];
  Sum[wickorder[i, x], {i, 0, m}]
];
  
(* VEV calculation using the Wick theorem *)  

sneglinearoperator[vevwick];
vevwick[HoldPattern[x : nc[l__]]] := wickorder[maxcran[x], x];
vevwick[] := 1;



EndPackage[];

"sneg.m $Id: sneg-Flat-OneIdentity.m,v 1.1 2006/07/04 16:13:22 rok Exp rok $ loaded"
